"""Rule to set/unset sleep state."""
import logging

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.interface
import HABApp.openhab.items
import HABApp.util

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class Sleep(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Rules class to manage sleep state.

	Example OpenHAB configuration:
	# KNX-things:
	Thing device T00_99_OpenHab_Sleep "KNX OpenHAB Sleep"{
        Type switch             : sleep             "Sleep Request"             [ ga="0/2/30"]
        Type switch-control     : sleep_RM          "Sleep RM"                  [ ga="0/2/31"]

        Type switch             : sleep_lock        "Sleep Lock Request"        [ ga="0/2/32"]
        Type switch-control     : sleep_lock_RM     "Sleep Lock RM"             [ ga="0/2/33"]

        Type string-control     : sleep_text        "Sleep Text"                [ ga="16.000:0/2/34"]
    }

    # Items:
    Switch    I01_02_Sleep              "Sleep [%s]"                <moon>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_RM"}
	Switch    I01_02_Sleep_req          "Sleep request"             <moon>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep"}
	String    I01_02_Sleep_text         "Text for display [%s]"                {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_text"}
	Switch    I01_02_Sleep_lock         "Lock [%s]"                 <lock>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_lock_RM"}
	Switch    I01_02_Sleep_lock_req     "Lock request"              <lock>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_lock"}
	String    I01_02_Sleep_State        "State [%s]"                <state>

	# Rule init:
	habapp_rules.system.sleep.Sleep("I01_02_Sleep","I01_02_Sleep_req", "I01_02_Sleep_State", "I01_02_Sleep_lock", "I01_02_Sleep_lock_req", "I01_02_Sleep_text")
	"""

	states = [
		{"name": "awake"},
		{"name": "pre_sleeping", "timeout": 3, "on_timeout": "pre_sleeping_timeout"},
		{"name": "sleeping"},
		{"name": "post_sleeping", "timeout": 3, "on_timeout": "post_sleeping_timeout"},
		{"name": "locked"},
	]

	trans = [
		{"trigger": "start_sleeping", "source": "awake", "dest": "pre_sleeping"},
		{"trigger": "pre_sleeping_timeout", "source": "pre_sleeping", "dest": "sleeping"},
		{"trigger": "end_sleeping", "source": "sleeping", "dest": "post_sleeping"},
		{"trigger": "end_sleeping", "source": "pre_sleeping", "dest": "awake", "unless": "lock_request_active"},
		{"trigger": "end_sleeping", "source": "pre_sleeping", "dest": "locked", "conditions": "lock_request_active"},
		{"trigger": "post_sleeping_timeout", "source": "post_sleeping", "dest": "awake", "unless": "lock_request_active"},
		{"trigger": "post_sleeping_timeout", "source": "post_sleeping", "dest": "locked", "conditions": "lock_request_active"},
		{"trigger": "set_lock", "source": "awake", "dest": "locked"},
		{"trigger": "release_lock", "source": "locked", "dest": "awake"}
	]

	def __init__(self, name_sleep: str, name_sleep_request: str, name_lock: str = None, name_lock_request: str = None, name_display_text: str = None, state_name: str = None, state_label: str = "Sleep state") -> None:
		"""Init of Sleep object.

		:param name_sleep: name of OpenHAB sleep item (SwitchItem)
		:param name_sleep_request: name of OpenHAB sleep request item (SwitchItem)
		:param name_lock: name of OpenHAB lock item (SwitchItem)
		:param name_lock_request: name of OpenHAB lock request item (SwitchItem)
		:param name_display_text: name of OpenHAB display text item (StringItem)
		:param state_name: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, state_name, state_label)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_sleep)

		# init items
		self.__item_sleep = HABApp.openhab.items.SwitchItem.get_item(name_sleep)
		self.__item_sleep_request = HABApp.openhab.items.SwitchItem.get_item(name_sleep_request)
		self.__item_lock = HABApp.openhab.items.SwitchItem.get_item(name_lock) if name_lock else None
		self.__item_lock_request = HABApp.openhab.items.SwitchItem.get_item(name_lock_request) if name_lock_request else None
		self.__item_display_text = HABApp.openhab.items.StringItem.get_item(name_display_text) if name_display_text else None

		# init attributes
		self._sleep_request_active = bool(self.__item_sleep_request)
		self._lock_request_active = bool(self.__item_lock_request) if self.__item_lock_request is not None else False

		# init state machine
		self.state_machine = habapp_rules.core.state_machine_rule.StateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self._update_openhab_state()

		# add callbacks
		self.__item_sleep_request.listen_event(self._cb_sleep_request, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self.__item_lock_request is not None:
			self.__item_lock_request.listen_event(self._cb_lock_request, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._instance_logger.debug(super().get_initial_log_message())

	def _get_initial_state(self, default_value: str = "awake") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: return correct state if it could be detected, if not return default value
		"""
		sleep_req = bool(self.__item_sleep_request) if self.__item_sleep_request.value is not None else None
		lock_req = bool(self.__item_lock_request) if self.__item_lock_request is not None and self.__item_lock_request.value is not None else None

		if sleep_req:
			return "sleeping"
		if lock_req:
			return "locked"
		if sleep_req is False:
			return "awake"

		return default_value

	@property
	def sleep_request_active(self) -> bool:
		"""Check if a sleep request is active

		:return: return true if lock request is active
		"""
		return self._sleep_request_active

	@property
	def lock_request_active(self) -> bool:
		"""Check if a lock request is active

		:return: return true if lock request is active
		"""
		return self._lock_request_active

	def _update_openhab_state(self):
		"""Extend _update_openhab state of base class to also update other OpenHAB items."""
		super()._update_openhab_state()

		# update sleep state
		if self.state in {"pre_sleeping", "sleeping"}:
			habapp_rules.core.helper.send_if_different(self.__item_sleep.name, "ON")
		else:
			habapp_rules.core.helper.send_if_different(self.__item_sleep.name, "OFF")

		# update lock state
		self.__update_lock_state()

		# update display text
		if self.__item_display_text is not None:
			self.__item_display_text.oh_send_command(self.__get_display_text())

	def __get_display_text(self) -> str:
		"""Get Text for displays.

		:return: display text
		"""
		if self.state == "awake":
			return "Schlafen"
		if self.state == "pre_sleeping":
			return "Guten Schlaf"
		if self.state == "sleeping":
			return "Aufstehen"
		if self.state == "post_sleeping":
			return "Guten Morgen"
		if self.state == "locked":
			return "Gesperrt"
		return ""

	def __update_lock_state(self):
		"""Update the return lock state value of OpenHAB item."""
		if self.__item_lock is not None:
			if self.state in {"pre_sleeping", "post_sleeping", "locked"}:
				habapp_rules.core.helper.send_if_different(self.__item_lock.name, "ON")
			else:
				habapp_rules.core.helper.send_if_different(self.__item_lock.name, "OFF")

	def _cb_sleep_request(self, event: HABApp.openhab.events.ItemStateChangedEvent):
		"""Callback, which is called if sleep request item changed state.

		:param event: Item state change event of sleep_request item
		"""
		if event.value == "ON" and self.state == "awake":
			self._instance_logger.debug("Start sleeping through sleep switch")
			self._sleep_request_active = True
			self.start_sleeping()
		elif event.value == "ON" and self.state == "locked":
			self._sleep_request_active = False
			self.__item_sleep_request.oh_send_command("OFF")
		elif event.value == "OFF" and self.state in {"sleeping", "pre_sleeping"}:
			self._instance_logger.debug("End sleeping through sleep switch")
			self._sleep_request_active = True
			self.end_sleeping()

	def _cb_lock_request(self, event: HABApp.openhab.events.ItemStateChangedEvent):
		"""Callback, which is called if lock request item changed state.

		:param event: Item state change event of sleep_request item
		"""
		self._lock_request_active = event.value == "ON"

		if self.state == "awake" and event.value == "ON":
			self.set_lock()
		elif self.state == "locked" and event.value == "OFF":
			self.release_lock()
		else:
			self.__update_lock_state()



"""Common helper functions for all rules."""

import HABApp.openhab.items


def send_if_different(item_name: str, value: str) -> None:
	"""Send command if the target value is different to the current value.

	:param item_name: name of OpenHab item
	:param value: value to write to OpenHAB item
	"""
	if str(HABApp.openhab.items.OpenhabItem.get_item(item_name).value) != value:
		HABApp.openhab.items.OpenhabItem.get_item(item_name).oh_send_command(value)



"""Setup logger."""
import collections.abc
import logging
import os
import typing

import HABApp.config.config
import HABApp.config.logging

import habapp_rules.__version__

LOG_PATH = HABApp.config.config.CONFIG.directories.logging.absolute()


def setup_logger() -> None:
	"""Setup the logger"""
	log_formatter = logging.Formatter("%(asctime)s.%(msecs)03d | %(threadName)20s | %(levelname)8s | %(name)s:%(lineno)d | %(message)s", datefmt="%Y-%m-%d | %H:%M:%S")
	habapp_rules_logger = logging.getLogger("habapp_rules")
	habapp_rules_logger.setLevel(logging.DEBUG)

	console_handler = logging.StreamHandler()
	console_handler.setFormatter(log_formatter)
	console_handler.setLevel(logging.DEBUG)
	habapp_rules_logger.addHandler(console_handler)

	if not LOG_PATH.is_dir():
		os.makedirs(LOG_PATH)

	file_handler = HABApp.config.logging.MidnightRotatingFileHandler(LOG_PATH / "habapp_rules.log", encoding="utf-8")
	file_handler.setFormatter(log_formatter)
	file_handler.setLevel(logging.DEBUG)
	habapp_rules_logger.addHandler(file_handler)


class InstanceLogger(logging.LoggerAdapter):
	"""Logging adapter to add the instance name to the log message."""

	def __init__(self, logger: logging.Logger, instance_name: str) -> None:
		"""Instantiate a logging adapter for multiple instances.

		:param logger: the underlying logger e.g. module logger
		:param instance_name: the name of the instance
		"""
		self._instance_name = instance_name
		logging.LoggerAdapter.__init__(self, logger)

	def process(self, msg: str, kwargs: collections.abc.MutableMapping[str, typing.Any]) -> tuple[str, collections.abc.MutableMapping[str, typing.Any]]:
		"""Add the instance name to log message.

		:param msg: the log message
		:param kwargs: additional keyword arguments.
		:return: tuple of msg with given keyword arguments
		"""
		return f"{self._instance_name} | {msg}", kwargs


setup_logger()
LOGGER = logging.getLogger(__name__)
LOGGER.info(f"Start logging of habapp_rules. Version = {habapp_rules.__version__.__version__}")



"""Module for hysteresis switch"""


class HysteresisSwitch:
	"""Hysteresis switch"""

	def __init__(self, threshold_on: float, hysteresis: float, return_bool: bool = True):
		"""Switch with hysteresis
		:param threshold_on: threshold for switching on
		:param hysteresis: hysteresis offset: threshold_off = threshold_on -hysteresis_offset
		:param return_bool: choose return-type: if true bool will be returned, else 'ON' / 'OFF'
		"""
		self._threshold = threshold_on
		self._hysteresis = hysteresis
		self._return_bool = return_bool
		self._on_off_state = False
		self._value_last = 0

	def set_threshold_on(self, threshold_on: float) -> None:
		"""Update threshold.

		:param threshold_on: new threshold value
		"""
		self._threshold = threshold_on

	def get_output(self, value: float | None = None) -> bool | str:
		"""Get output of hysteresis switch
		:param value: value which should be checked
		:return: on / off state
		"""
		# get threshold depending on the current state
		threshold = self._threshold - 0.5 * self._hysteresis if self._on_off_state else self._threshold + 0.5 * self._hysteresis

		# use new value if given, otherwise last value
		value = value if value else self._value_last

		# save value for next check
		self._value_last = value

		# get on / off state
		self._on_off_state = value >= threshold

		# if on/off result is requested convert result
		if self._return_bool:
			return self._on_off_state

		return "ON" if self._on_off_state else "OFF"



"""Rule to detect presence or absence."""
import logging
import threading

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.interface
import HABApp.openhab.items
import HABApp.util

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class Presence(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Rules class to manage presence of a home.

	Hint: If you have some kind of guest-mode, use a guest-available switch as a phone to enable a persistent presence, also if all phones are not at home

	Example OpenHAB configuration:
	# KNX-things:
	Thing device T00_99_OpenHab_Presence "KNX OpenHAB Presence"{
        Type switch-control        : presence       "Presence"      [ ga="0/2/11+0/2/10"]
        Type switch-control        : leaving        "Leaving"       [ ga="0/2/21+0/2/20"]
    }

    # Items:
    Switch    I01_00_Presence    "Presence [%s]"    <presence>    (G00_00_rrd4j)    ["Status", "Presence"]    {channel="knx:device:bridge:T00_99_OpenHab_Presence:presence"}
	Switch    I01_00_Leaving     "Leaving [%s]"     <leaving>                                                 {channel="knx:device:bridge:T00_99_OpenHab_Presence:leaving"}

	# Rule init:
	habapp_rules.system.presence.Presence("I01_00_Presence", "I01_00_Leaving")
	"""

	states = [
		{"name": "presence"},
		{"name": "leaving", "timeout": 5 * 60, "on_timeout": "absence_detected"},  # leaving takes 5 minutes
		{"name": "absence", "timeout": 1.5 * 24 * 3600, "on_timeout": "long_absence_detected"},  # switch to long absence after 1.5 days
		{"name": "long_absence"}
	]

	trans = [
		{"trigger": "presence_detected", "source": ["absence", "long_absence"], "dest": "presence"},
		{"trigger": "leaving_detected", "source": ["presence", "absence", "long_absence"], "dest": "leaving"},
		{"trigger": "abort_leaving", "source": "leaving", "dest": "presence"},
		{"trigger": "absence_detected", "source": ["presence", "leaving"], "dest": "absence"},
		{"trigger": "long_absence_detected", "source": "absence", "dest": "long_absence"},
	]

	def __init__(self, name_presence: str, leaving_name: str, outside_door_names: list[str] | None = None, phone_names: list[str] | None = None, state_name: str | None = None, state_label: str = "Presence state") -> None:
		"""Init of Presence object.

		:param name_presence: name of OpenHAB presence item
		:param leaving_name: name of OpenHAB leaving item (SwitchItem)
		:param outside_door_names: list of names of OpenHAB outdoor door items
		:param phone_names: list of names of OpenHAB phone items
		:param state_name: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""
		if not outside_door_names:
			outside_door_names = []
		if not phone_names:
			phone_names = []

		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, state_name, state_label)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_presence)

		# init items
		self.__presence_item = HABApp.openhab.items.SwitchItem.get_item(name_presence)
		self.__leaving_item = HABApp.openhab.items.SwitchItem.get_item(leaving_name)
		self.__outside_door_items = [HABApp.openhab.items.ContactItem.get_item(name) for name in outside_door_names]
		self.__phone_items = [HABApp.openhab.items.SwitchItem.get_item(name) for name in phone_names]

		# init state machine
		self.state_machine = habapp_rules.core.state_machine_rule.StateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		# add callbacks
		self.__leaving_item.listen_event(self._cb_leaving, HABApp.openhab.events.ItemStateChangedEventFilter())
		self.__presence_item.listen_event(self._cb_presence, HABApp.openhab.events.ItemStateChangedEventFilter())
		HABApp.util.EventListenerGroup().add_listener(self.__outside_door_items, self._cb_outside_door, HABApp.core.events.ValueChangeEventFilter()).listen()
		HABApp.util.EventListenerGroup().add_listener(self.__phone_items, self._cb_phone, HABApp.core.events.ValueChangeEventFilter()).listen()

		self.__phone_absence_timer: threading.Timer | None = None
		self._instance_logger.debug(super().get_initial_log_message())

	def _get_initial_state(self, default_value: str = "presence") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: return correct state if it could be detected, if not return default value
		"""
		phone_items = [phone for phone in self.__phone_items if phone.value is not None]
		if phone_items:
			if any((item.value == "ON" for item in phone_items)):
				return "presence"

			if self.__presence_item.value == "ON":
				return "leaving"
			return "long_absence" if self._item_state.value == "long_absence" else "absence"

		if self.__leaving_item.value == "ON":
			return "leaving"

		if self.__presence_item.value == "ON":
			return "presence"

		if self.__presence_item.value == "OFF":
			return "long_absence" if self._item_state.value == "long_absence" else "absence"

		return default_value

	def _update_openhab_state(self) -> None:
		"""Extend _update_openhab state of base class to also update other OpenHAB items."""
		super()._update_openhab_state()
		self._instance_logger.info(f"Presence state changed to {self.state}")

		# update presence item
		target_value = "ON" if self.state in {"presence", "leaving"} else "OFF"
		habapp_rules.core.helper.send_if_different(self.__presence_item.name, target_value)

		habapp_rules.core.helper.send_if_different(self.__leaving_item.name, "ON" if self.state == "leaving" else "OFF")

	def _cb_outside_door(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if any outside door changed state.

		:param event: state change event of door item
		"""
		if event.value == "OPEN" and self.state != "presence":
			self._instance_logger.debug(f"Presence detected by door ({event.name})")
			self.presence_detected()

	def _cb_leaving(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if leaving item changed state.

		:param event: Item state change event of leaving item
		"""
		if event.value == "ON" and self.state in {"presence", "absence", "long_absence"}:
			self._instance_logger.debug("Start leaving through leaving switch")
			self.leaving_detected()
		if event.value == "OFF" and self.state == "leaving":
			self._instance_logger.debug("Abort leaving through leaving switch")
			self.abort_leaving()

	def _cb_presence(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if presence item changed state.

		:param event: Item state change event of presence item
		"""
		if event.value == "ON" and self.state in {"absence", "long_absence"}:
			self._instance_logger.debug("Presence was set manually by presence switch")
			self.presence_detected()
		elif event.value == "OFF" and self.state in {"presence", "leaving"}:
			self._instance_logger.debug("Absence was set manually by presence switch")
			self.absence_detected()

	def _cb_phone(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if a phone state changed.

		:param event: Item state change event of phone item
		"""
		active_phones = len([phone for phone in self.__phone_items if phone.value == "ON"])
		if active_phones == 1 and event.value == "ON":
			# first phone switched to ON
			if self.__phone_absence_timer:
				self.__phone_absence_timer.cancel()
				self.__phone_absence_timer = None

			if self.state in {"absence", "long_absence"}:
				self._instance_logger.debug("Presence was set through first phone joined network")
				self.presence_detected()

		elif active_phones == 0 and event.value == "OFF":
			# last phone switched to OFF
			self.__phone_absence_timer = threading.Timer(20 * 60, self.__set_leaving_through_phone)
			self.__phone_absence_timer.start()

	def __set_leaving_through_phone(self) -> None:
		"""Set leaving detected if timeout expired."""
		if self.state == "presence":
			self._instance_logger.debug("Leaving was set, because last phone left some time ago.")
			self.leaving_detected()
		self.__phone_absence_timer = None



import typing

import setuptools

import habapp_rules.__version__


def load_req() -> typing.List[str]:
	with open('requirements.txt') as f:
		return f.readlines()


VERSION = habapp_rules.__version__.__version__

setuptools.setup(
	name="habapp_rules",
	version=VERSION,
	author="Seuling N.",
	description="Basic rules for HABApp",
	long_description="Basic rules for HABApp",
	packages=setuptools.find_packages(exclude=["tests*", "rules*"]),
	install_requires=load_req(),
	python_requires=">=3.10",
	license="Apache License 2.0"
)
